#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import numpy as np


from .. import representations
from .mappings import coding_maps

RBalgo = representations.RBalgo


def ZPKrep2MRF(
    ZPKrep,
    coding_map=coding_maps.nlFBW_safe,
    delay_s=None,
    insert_poles=None,
    insert_zeros=None,
    distance_limit_auto=None,
    **kwargs
):
    try:
        repdict = coding_map.rep
    except AttributeError:
        coding_submap = coding_map
    else:
        try:
            if ZPKrep.F_nyquist_Hz is None:
                rep = "Sf"
            else:
                rep = "Z"
            coding_submap = repdict[rep]
        except KeyError:
            raise RuntimeError(
                (
                    'Conversion to representation "{}" not supported'
                    " in used coding_map."
                ).format(ZPKrep.rep)
            )

    gain_coding = coding_submap.gain()
    gain_coding.setup(gain=ZPKrep.gain)

    if delay_s is None:
        delay_s = ZPKrep.delay_s

    # pregenerate the fitter to use as the system source for the codings
    fitterMRF = coding_submap.mrf_default(
        data=ZPKrep.data,
        W=ZPKrep.W,
        F_Hz=ZPKrep.F_Hz,
        zeros_overlay=ZPKrep.zeros_overlay,
        poles_overlay=ZPKrep.poles_overlay,
        F_nyquist_Hz=ZPKrep.F_nyquist_Hz,
        residuals_log_im_scale=ZPKrep.residuals_log_im_scale,
        delay_s=delay_s,
        num_codings=[],
        den_codings=[],
        gain_coding=gain_coding,
        coding_map=coding_submap,
        distance_limit_auto=distance_limit_auto,
        **kwargs
    )
    fitterMRF.delay_s = ZPKrep.delay_s

    # TODO, reset the gain to use the gain_effect parameters
    num_codings = []
    den_codings = []
    Z = RBalgo.expect(
        ZPKrep.zeros, constraint=representations.root_constraints.mirror_real
    )
    P = RBalgo.expect(
        ZPKrep.poles, constraint=representations.root_constraints.mirror_real
    )
    if insert_zeros is not None:
        # TODO, add completion?
        Z = Z * RBalgo.expect(
            insert_zeros, constraint=representations.root_constraints.mirror_real
        )
    if insert_poles is not None:
        # TODO, add completion?
        P = P * RBalgo.expect(
            insert_poles, constraint=representations.root_constraints.mirror_real
        )

    num_codings = []
    num_codings.extend(
        encode(
            fitterMRF,
            Z.r,
            coding_submap.num_r,
            coding_submap.num_r_u,
            coding_submap.num_collect_r,
            coding_submap.is_unstable,
        )
    )
    num_codings.extend(
        encode(
            fitterMRF,
            Z.c,
            coding_submap.num_c,
            coding_submap.num_c_u,
            coding_submap.num_collect_c,
            coding_submap.is_unstable,
        )
    )

    den_codings = []
    den_codings.extend(
        encode(
            fitterMRF,
            P.r,
            coding_submap.den_r,
            coding_submap.den_r_u,
            coding_submap.den_collect_r,
            coding_submap.is_unstable,
        )
    )
    den_codings.extend(
        encode(
            fitterMRF,
            P.c,
            coding_submap.den_c,
            coding_submap.den_c_u,
            coding_submap.den_collect_c,
            coding_submap.is_unstable,
        )
    )

    # now set the codings
    fitterMRF.num_codings = num_codings
    fitterMRF.den_codings = den_codings
    # set the gain again to account for gain_effect
    fitterMRF.gain = ZPKrep.gain
    # TODO, HACK
    return fitterMRF


def encode(fitter, r_list, st_t, us_t, collect, is_unstable):
    """
    Internal function to perform the encoding
    """
    if us_t is not None:
        r_list_st = []
        r_list_us = []
        for root in r_list:
            if is_unstable(root):
                r_list_us.append(root)
            else:
                r_list_st.append(root)
    else:
        r_list_st = list(r_list)
        r_list_us = []

    if us_t is None:
        us_t = st_t

    codings = []
    while r_list_st:
        coding = st_t(fitter)
        roots = r_list_st[-collect:]
        r_list_st[-collect:] = []
        coding.update_roots(*roots)
        codings.append(coding)
        rts = coding.roots()
        rts = np.asarray(rts)
        assert(not rts.size or np.all(np.isfinite(rts)))
    while r_list_us:
        coding = us_t(fitter)
        roots = r_list_us[-collect:]
        r_list_us[-collect:] = []
        coding.update_roots(*roots)
        rts = coding.roots()
        rts = np.asarray(rts)
        assert(not rts.size or np.all(np.isfinite(rts)))
        codings.append(coding)
    return codings


def MRF2MRF(
    fitter,
    ZPKrep=None,
    zeros=None,
    poles=None,
    gain=None,
    p_c=None,
    z_c=None,
    p_r=None,
    z_r=None,
    delay_s=None,
    coding_map=None,
    **kwargs
):

    if ZPKrep is None:
        ZPKrep = fitter.ZPKrep

    if zeros is None:
        zeros = ZPKrep.zeros
    else:
        zeros = fitter.RBalgo.expect(zeros)

    if poles is None:
        poles = ZPKrep.poles
    else:
        poles = fitter.RBalgo.expect(poles)

    if z_r is None:
        z_r = zeros.r
    else:
        z_r = np.asarray(z_r)

    if p_r is None:
        p_r = poles.r
    else:
        p_r = np.asarray(p_r)

    if z_c is None:
        z_c = zeros.c
    else:
        z_c = np.asarray(z_c)

    if p_c is None:
        p_c = poles.c
    else:
        p_c = np.asarray(p_c)

    if gain is None:
        gain = ZPKrep.gain

    assert(not z_r.size or np.all(np.isfinite(z_r)))
    assert(not z_c.size or np.all(np.isfinite(z_c)))
    assert(not p_r.size or np.all(np.isfinite(p_r)))
    assert(not p_c.size or np.all(np.isfinite(p_c)))
    zeros = representations.asMRRB(r=z_r, c=z_c)
    poles = representations.asMRRB(r=p_r, c=p_c)
    coding_map, num_codings, den_codings = ZP2codings(
        fitter, zeros, poles, coding_map=coding_map
    )

    gain_coding = coding_map.gain()
    gain_coding.setup(gain=gain)

    if delay_s is None:
        delay_s = ZPKrep.delay_s

    # pregenerate the fitter to use as the system source for the codings
    fitterMRF = coding_map.mrf_default(
        parent=fitter,
        data=ZPKrep.data,
        F_Hz=ZPKrep.F_Hz,
        W=ZPKrep.W,
        num_codings=num_codings,
        den_codings=den_codings,
        gain_coding=gain_coding,
        codings=coding_map.module,
        zeros_overlay=ZPKrep.zeros_overlay,
        poles_overlay=ZPKrep.poles_overlay,
        delay_s=delay_s,
        **kwargs
    )
    return fitterMRF


def ZP2codings(fitter, zeros, poles, coding_map=None, **kwargs):
    if coding_map is None:
        coding_map = fitter.coding_map

    try:
        repdict = coding_map.rep
    except AttributeError:
        coding_map = coding_map
    else:
        if fitter.F_nyquist_Hz is None:
            rep = "Sf"
        else:
            rep = "Z"
        coding_map = repdict[rep]

    zeros = fitter.RBalgo.expect(zeros, fitter.root_constraint)
    poles = fitter.RBalgo.expect(poles, fitter.root_constraint)

    num_codings = []
    num_codings.extend(
        encode(
            fitter,
            zeros.r,
            coding_map.num_r,
            coding_map.num_r_u,
            coding_map.num_collect_r,
            coding_map.is_unstable,
        )
    )
    num_codings.extend(
        encode(
            fitter,
            zeros.c,
            coding_map.num_c,
            coding_map.num_c_u,
            coding_map.num_collect_c,
            coding_map.is_unstable,
        )
    )
    den_codings = []
    den_codings.extend(
        encode(
            fitter,
            poles.r,
            coding_map.den_r,
            coding_map.den_r_u,
            coding_map.den_collect_r,
            coding_map.is_unstable,
        )
    )
    den_codings.extend(
        encode(
            fitter,
            poles.c,
            coding_map.den_c,
            coding_map.den_c_u,
            coding_map.den_collect_c,
            coding_map.is_unstable,
        )
    )

    return coding_map, num_codings, den_codings



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from .any_io import (
    load_any,
    write_any,
)

from .csv_io import (
    load_csv,
)

from .types import (
    type2type,
    ext2type,
    type2features,
    determine_type,
)

from .utilities import (
    subkey_search,
)


def save(fname, d):
    typeB = determine_type(fname)
    write_any(
        fname=typeB.fname,
        ftype=typeB.ftype,
        fdict=d,
    )


def load(fname, ftype=None):
    typeB = determine_type(fname)
    if ftype is None:
        ftype = typeB.ftype

    return load_any(
        fname=typeB.fname,
        ftype=ftype,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from .TF import (
    TF_ZPK,
    ZtoS,
    StoZ,
    SorZtoSorZ,
    ZPK_fill,
)

from .roots_bin import (
    roots_bin_type,
    roots_re_pair,
)


def abs_sq(x):
    return x.real ** 2 + x.imag ** 2


def norm_sq(v):
    return v.real.dot(v.real) + v.imag.dot(v.imag)


from numpy.polynomial.polynomial import polyvalfromroots

from .roots_matching import (
    nearest_idx,
    nearest_unique_idx,
    nearest_pairs,
    nearest_unique_pairs,
    SOS_pair_rolloff,
    match_SOS_pairs,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import argparse
from ...utilities.strings import padding_remove


def kwdict_argparse(ap, kwdict, groups_kw=dict()):
    groups = dict()
    groups_prio = dict()
    for gname, gdict in groups_kw.items():
        APpriority = gdict.get("APpriority", None)
        if APpriority is not None:
            groups_prio[gname] = APpriority

    for hname, hdict in kwdict.items():
        if hdict.get("APignore", False):
            continue
        group = hdict.get("APgroup", None)
        s = groups.setdefault(group, set())
        s.add(hname)

        gprio = groups_prio.get(group, float("inf"))
        hprio = hdict.get("APpriority", float("inf"))
        if hprio < gprio:
            gprio = hprio

        groups_prio[group] = gprio

    gnames = list(groups_prio.keys())
    gnames.sort(
        key=lambda g: (groups_prio[g], g)
        if groups_prio[g] is not None
        else float("inf")
    )

    for gname in gnames:
        hset = groups[gname]
        gkw = dict()
        if gname is None:
            apG = ap
        else:
            gdict = groups_kw.get(gname, None)
            if gdict is not None:
                ghelp = gdict.get("about", None)
                if ghelp is not None:
                    gkw["description"] = padding_remove(ghelp)
        apG = ap.add_argument_group(gname, **gkw)
        hlist = list(hset)
        hlist.sort(key=lambda h: (kwdict[h].get("APpriority", float("inf")), h))
        for hname in hlist:
            hdict = kwdict[hname]
            name = hdict.get(hname, hname)

            APshort = hdict.get("APshort", None)
            APlong = "--{}".format(name)
            APflags = list(hdict.get("APflags", [APlong]))

            if APshort is not None:
                APflags = [APshort] + APflags

            APhide = hdict.get("APhide", False)
            if APhide:
                helptext = argparse.SUPPRESS
            else:
                helptext = hdict.get("about", "<needs doc>").replace("%", "%%")
                helptext = padding_remove(helptext)
            APkw = dict(
                dest=name,
                help=helptext,
            )

            APaction = hdict.get("APaction", None)
            if APaction is not None:
                APkw["action"] = APaction

            APtype = hdict.get("APtype", None)
            if APtype is not None:
                APkw["type"] = APtype

            APrequired = hdict.get("APrequired", None)
            if APrequired is not None:
                APkw["required"] = APrequired

            APnargs = hdict.get("APnargs", None)
            if APnargs is not None:
                APkw["nargs"] = APnargs

            APchoices = hdict.get("APchoices ", None)
            if APchoices is not None:
                APkw["choices "] = APchoices

            APmetavar = hdict.get("APmetavar", None)
            if APmetavar is not None:
                APkw["metavar"] = APmetavar

            APchoices = hdict.get("APchoices", None)
            if APchoices is not None:
                APkw["choices"] = APchoices

            APconst = hdict.get("APconst", None)
            if APconst is not None:
                APkw["const"] = APconst

            APdefault = hdict.get("APdefault", None)
            if APdefault is not None:
                APkw["default"] = APdefault
            else:
                APkw["default"] = argparse.SUPPRESS

            APpositional = hdict.get("APpositional", False)

            if not APpositional:
                apG.add_argument(*APflags, **APkw)

                # add the aliases to parse, but with their help suppressed
                for aname in hdict.get("aliases", []):
                    APkw["help"] = argparse.SUPPRESS
                    apG.add_argument("--{}".format(aname), **APkw)
                for aname in hdict.get("aliases_bad", []):
                    APkw["help"] = argparse.SUPPRESS
                    apG.add_argument("--{}".format(aname), **APkw)
            else:
                dest = APkw.pop("dest")
                apG.add_argument(dest, **APkw)

    return ap



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


def ini_load(fname):
    import configparser

    dummy_section = "dummy_section"
    with open(fname) as f:
        file_content = "[{}]\n".format(dummy_section) + f.read()

    cp = configparser.ConfigParser(interpolation=None)
    cp.read_string(file_content)
    base = dict()
    for op in cp.options(dummy_section):
        base[op] = cp.get(dummy_section, op)
    sections = dict()
    for sec in cp.sections():
        d = dict()
        sections[sec] = d
        for op in cp.options(sec):
            d[op] = cp.get(sec, op)
    if set(base.keys()) ^ set(sections.keys()):
        raise RuntimeError(
            ("ini file {} has setting which conflicts with section name").format(fname)
        )
    base.update(sections)
    return base



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

from .utils import *

from .autoniceplot import (
    AutoPlotSaver,
    asavefig,
    mplfigB,
)

from .stacked_plots import (
    generate_stacked_plot_ax,
    generate_ax,
)

from .logticks import (
    setup_log_xticks,
)

style_6p5in = {
    "savefig.dpi": 100,
    "figure.dpi": 100,
    "lines.linewidth": 1.0,
    "axes.facecolor": "FFFFFF",
    "figure.facecolor": "FFFFFF",
    "figure.figsize": [6.5, 3.0],
    "font.family": "DejaVu Sans",
    "figure.titlesize": "large",  # size of the figure title (Figure.suptitle())
    "figure.titleweight": "normal",  # weight of the figure title
    "pdf.fonttype": 3,  # Output Type 3 (Type3) or Type 42 (TrueType)
    "font.size": 10.0,
    "axes.grid": False,  # display grid or not
    "axes.titlesize": "large",  # fontsize of the axes title
    "axes.titlepad": 6.0,  # pad between axes and title in points
    "axes.labelsize": "medium",  # fontsize of the x any y labels
    "axes.labelpad": 4.0,  # space between label and axis
    "axes.formatter.useoffset": True,  # If True, the tick label formatter
    "axes.formatter.offset_threshold": 4,  # When useoffset is True, the offset
    "legend.fontsize": "medium",
    "legend.borderpad": 0.4,  # border whitespace
    "legend.labelspacing": 0.5,  # the vertical space between the legend entries
    "legend.handlelength": 2.0,  # the length of the legend lines
    "legend.handleheight": 0.7,  # the height of the legend handle
    "legend.handletextpad": 0.8,  # the space between the legend line and legend text
    "legend.borderaxespad": 0.5,  # the border between the axes and legend edge
    "legend.columnspacing": 2.0,  # column separation
}


style_3p5in = {
    "savefig.dpi": 200,
    "figure.dpi": 200,
    "lines.linewidth": 1.0,
    "axes.facecolor": "FFFFFF",
    "figure.facecolor": "FFFFFF",
    "figure.figsize": [3.5, 2.5],
    "font.family": "DejaVu Serif",
    "figure.titlesize": "medium",  # size of the figure title (Figure.suptitle())
    "figure.titleweight": "normal",  # weight of the figure title
    "pdf.fonttype": 3,  # Output Type 3 (Type3) or Type 42 (TrueType)
    "font.size": 10.0,
    "axes.grid": False,  # display grid or not
    "axes.titlesize": 10,  # fontsize of the axes title
    "axes.titlepad": 6.0,  # pad between axes and title in points
    "axes.labelsize": 8,  # fontsize of the x any y labels
    "axes.labelpad": 4.0,  # space between label and axis
    "axes.formatter.useoffset": True,  # If True, the tick label formatter
    "axes.formatter.offset_threshold": 4,  # When useoffset is True, the offset
    "legend.fontsize": 8,
    "legend.borderpad": 0.4,  # border whitespace
    "legend.labelspacing": 0.5,  # the vertical space between the legend entries
    "legend.handlelength": 2.0,  # the length of the legend lines
    "legend.handleheight": 0.7,  # the height of the legend handle
    "legend.handletextpad": 0.8,  # the space between the legend line and legend text
    "legend.borderaxespad": 0.5,  # the border between the axes and legend edge
    "legend.columnspacing": 2.0,  # column separation
}

style_tex_serif = {
    "text.latex.preamble": r"\usepackage{txfonts}",
    "font.family": "serif",
    "text.usetex": True,
}

style_serif = {
    "font.family": "serif",
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np

from ..codings_cmn import (
    CodingTypeZ,
    Ipi,
)

# import scipy.linalg


class CodingFnlBW(CodingTypeZ):
    N_parameters = 2
    p_F_Hz = 0
    p_nl_BW = 0
    restabilize = False
    unstable = False

    def update(self, F_Hz, nl_BW):
        self.p_F_Hz = F_Hz
        self.p_nl_BW = nl_BW

    def reduce(self):
        return [
            self.p_F_Hz,
            self.p_nl_BW,
        ]

    def update_roots(self, r1):
        """ """
        amp = abs(r1)
        if self.restabilize:
            if amp > 1:
                print("stabilizing: ", r1)
                amp = 1 / amp
                self.lock_nl_BW = True
        self.p_F_Hz = np.angle(r1) / np.pi * self.sys.F_nyquist_Hz

        BW_Hz = (amp - 1) * self.sys.F_nyquist_Hz
        # TODO, use deadzoning instead of this method
        if BW_Hz <= 0:
            self.unstable = False
            self.p_nl_BW = (-BW_Hz) ** 0.5
        else:
            self.unstable = True
            self.p_nl_BW = BW_Hz ** 0.5
        return

    @property
    def F_Hz(self):
        return self.p_F_Hz

    @property
    def BW(self):
        if not self.unstable:
            BW = -self.p_nl_BW ** 2
        else:
            BW = self.p_nl_BW ** 2
        return BW

    def transfer(self):
        # frequency, linear nl_BW
        amp = 1 + self.BW / self.sys.F_nyquist_Hz

        r = amp * np.cos(np.pi * self.p_F_Hz / self.sys.F_nyquist_Hz)
        Xn = self.sys.Xzn_grid
        Xnsq = self.sys.Xzn_grid_sq
        return (amp * amp) * Xnsq - 2 * Xn * r + 1

    def derivative(self):
        if self.disable:
            return []
        # real/imaginary part of root
        if not self.unstable:
            DampDl = -(2 * self.p_nl_BW) / self.sys.F_nyquist_Hz
            amp = 1 - self.p_nl_BW / self.sys.F_nyquist_Hz
        else:
            DampDl = +(2 * self.p_nl_BW) / self.sys.F_nyquist_Hz
            amp = 1 + self.p_nl_BW / self.sys.F_nyquist_Hz

        F_nyquist_Hz = self.sys.F_nyquist_Hz
        Xn = self.sys.Xzn_grid
        Xnsq = self.sys.Xzn_grid_sq

        rcos = np.cos(np.pi * self.p_F_Hz / F_nyquist_Hz)
        rsin = np.sin(np.pi * self.p_F_Hz / F_nyquist_Hz)
        return [
            (2 * amp * np.pi / self.sys.F_nyquist_Hz) * (Xn * rsin),
            ((2 * amp * DampDl) * Xnsq - Xn * (2 * rcos * DampDl)),
        ]

    def roots_c(self):
        # frequency, linear nl_BW
        amp = 1 + self.BW / self.sys.F_nyquist_Hz
        return [amp * np.exp(Ipi * abs(self.p_F_Hz) / self.sys.F_nyquist_Hz)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ...fitters_rational import RationalDiscFilter, ChebychevFilter

from . import nyquist_move


def rational_disc_fit(
    ZPKrep,
    order=None,
    order_max=None,
    order_min=20,
    aid=None,
):
    nyquist_final_Hz = None

    if nyquist_final_Hz is not None and nyquist_final_Hz <= 0:
        raise RuntimeError(
            "nyquist_final_Hz must be positive for Z domain or None" " for Sf domain."
        )

    if order is not None:
        order = int(order)
    order_max = int(order_max)
    order_min = int(order_min)

    N_first = order_min
    N_final = int(min(len(ZPKrep.F_Hz) // 10, order_max))

    if order is not None:
        return ratdisc_single(
            ZPKrep,
            order=order,
            aid=aid,
        )

    # otherwise, scan through
    N_current = N_first

    fitter_last = ratdisc_single(
        ZPKrep,
        order=N_current,
        aid=aid,
    )
    restot_last = fitter_last.residuals_average

    def fitter_ord(fitter):
        return max(len(fitter.zeros), len(fitter.poles))

    while True:
        if N_current == N_final:
            fitter_use = fitter_last
            break

        N_current = N_current * 2
        if N_current > N_final:
            N_current = N_final

        fitter = ratdisc_single(
            ZPKrep,
            order=N_current,
            aid=aid,
        )
        restot = fitter.residuals_average
        fitter_red = fitter.copy()
        fitter_red.matched_pairs_clear(Q_rank_cutoff=0.2)
        restot_red = fitter_red.residuals_average

        if restot_last < restot:
            fitter_use = fitter_last
            aid.log("Using last (direct)!", fitter_ord(fitter_last))
            break

        if restot_last < restot_red:
            fitter_use = fitter_last
            aid.log("Using last (reduced)!", fitter_ord(fitter_last))
            break

        if restot_last < 1.10 * restot:
            ord_red = fitter_ord(fitter_red)
            ord_last = fitter_ord(fitter_last)
            if ord_red < ord_last:
                fitter_use = fitter_red
            else:
                fitter_use = fitter_last
            aid.log("Using current")
            break
        # else continue the loop
        fitter_last = fitter
        restot_last = restot

    fitter = fitter_use
    if nyquist_final_Hz == "keep":
        # this means to not perform the nyquist shift,
        # only really used in testing
        pass
    elif nyquist_final_Hz is None or nyquist_final_Hz == 0:
        ZPKrep = nyquist_move.nyquist_remove(
            fitter,
            split_neg_real=True,
            clear_neg_real=True,
            aid=aid,
        )
        fitter = ChebychevFilter(
            ZPKrep=ZPKrep,
        )
    elif nyquist_final_Hz > 0:
        fitter = nyquist_move.nyquist_move(
            fitter,
            nyquist_new=nyquist_final_Hz,
            split_neg_real=True,
            clear_neg_real=True,
            aid=aid,
        )
    else:
        raise RuntimeError(
            (
                "Nyquist cannot be negative. "
                " Positive for Z-domain, None or 0 for S."
                ' or "keep" to maintain the maxdata nyquist'
            )
        )

    return fitter


def ratdisc_single(
    ZPKrep,
    order=20,
    relative_degree=0,
    nyquist_final_Hz=None,
    aid=None,
):

    if nyquist_final_Hz is not None and nyquist_final_Hz <= 0:
        raise RuntimeError("nyquist_final_Hz must be >0 or None")

    F_nyquist_Hz = 1 * ZPKrep.F_Hz[-1]

    fitter = RationalDiscFilter(
        ZPKrep=ZPKrep,
        npoles=order,
        nzeros=order,
        F_nyquist_Hz=F_nyquist_Hz,
    )
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_SVD()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.5)
    fitter.stabilize = True
    # fitter.npoles = len(fitter.poles)
    # fitter.nzeros = len(fitter.zeros)
    ##fitter.stabilize = True
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_SVD()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.5)
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    return fitter



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import numpy as np
from wield.bunch import Bunch

from ... import TFmath
from ... import representations

from ...utilities import ensure_aid


def nyquist_move(
    fitter,
    nyquist_new,
    clear_neg_real=False,
    split_neg_real=False,
    aid=None,
):
    """
    Modifies a fitter IN PLACE to use a different nyquist frequency
    """
    # TODO, update this to properly use RootBunches and ZPKreps
    aid = ensure_aid(aid)
    ZPKrep = fitter.ZPKrep

    F_nyquist_Hz = ZPKrep.F_nyquist_Hz
    op = Bunch(gain_flip=1)

    # currently a restriction in the algorithm, it can't handle aliasing of poles (need bilinear for that)
    assert nyquist_new > F_nyquist_Hz
    ratio = nyquist_new / F_nyquist_Hz

    def root_list_map(root_list, rtype):
        new_list = []
        for root in root_list:
            new_roots = None
            if abs(root.imag) < 1e-15:
                # aid.log(root)
                if root.real > 0:
                    if root.real < 1:
                        new_roots = [1 - (1 - root.real) / ratio]
                    else:
                        # actually, just clear it
                        aid.log("Cleared ", rtype, root)
                        # new_roots = []
                        # flip it then insert
                        op.gain_flip *= -1
                        root = 1 / root
                        # new_roots = [1 - (1 - root.real) / ratio]
                        new_roots = []
                else:
                    if clear_neg_real:
                        aid.log("Cleared ", rtype, root)
                        new_roots = []
                    elif split_neg_real:
                        angle = np.pi / ratio
                        amp = -root.real
                        if amp < 1:
                            amp = 1 - (1 - amp) / ratio
                        else:
                            amp = (amp - 1) / ratio + 1
                        new_roots = [
                            amp * np.exp(1j * angle),
                            amp * np.exp(-1j * angle),
                        ]
                    else:
                        new_roots = [-1 - (-1 - root.real) / ratio]
            else:
                amp = abs(root)
                angle = np.angle(root)
                if angle > 0:
                    angle = angle / ratio
                else:
                    angle = angle / ratio
                if amp < 1:
                    amp = 1 - (1 - amp) / ratio
                else:
                    amp = (amp - 1) / ratio + 1
                new_roots = [amp * np.exp(1j * angle)]
            new_list.extend(new_roots)
        return new_list

    xfer_pre = fitter.xfer_fit

    nyquist_prev = ZPKrep.F_nyquist_Hz
    fitter.F_nyquist_Hz = nyquist_new
    poles = root_list_map(fitter.poles.fullplane, rtype="pole")
    zeros = root_list_map(fitter.zeros.fullplane, rtype="zero")
    fitter.poles = poles
    fitter.zeros = zeros

    # don't use the root mapper above since it does special things and can drop
    # overlays
    Zov, Pov, Kex = TFmath.SorZtoSorZ(
        (fitter.zeros_overlay.fullplane, fitter.poles_overlay.fullplane, 1),
        F_nyquist_Hz_in=nyquist_prev,
        F_nyquist_Hz_out=nyquist_new,
    )
    fitter.poles_overlay = Pov
    fitter.zeros_overlay = Zov

    xfer_post = fitter.xfer_fit

    gain_scale = np.median(abs(xfer_post / xfer_pre))

    gain = fitter.gain / gain_scale * op.gain_flip
    fitter.gain = gain
    # aid.log("GAIN: ", fitter.gain)
    return fitter


def nyquist_remove(
    fitter,
    clear_neg_real=False,
    split_neg_real=False,
    aid=None,
):
    """
    Modifies a fitter IN PLACE to use a different nyquist frequency
    """
    aid = ensure_aid(aid)

    F_nyquist_Hz = fitter.F_nyquist_Hz
    op = Bunch(gain_flip=1)

    def root_list_map(root_list, rtype):
        new_list = []
        for root in root_list:
            new_roots = None
            if abs(root.imag) < 1e-15:
                # aid.log(root)
                if root.real > 0:
                    if root.real < 1:
                        new_roots = [F_nyquist_Hz * (root.real - 1)]
                    else:
                        aid.log("Cleared ", rtype, root)
                        op.gain_flip *= -1
                        new_roots = []
                else:
                    if clear_neg_real:
                        aid.log("Cleared ", rtype, root)
                        # TODO, not sure about this one
                        op.gain_flip *= -1
                        new_roots = []
                    elif split_neg_real:
                        BW = -(1 + root.real) * F_nyquist_Hz
                        new_roots = [
                            BW + 1j * F_nyquist_Hz,
                            BW + -1j * F_nyquist_Hz,
                        ]
                    else:
                        new_roots = [F_nyquist_Hz * -(root.real + 1)]
            else:
                amp = abs(root)
                F_Hz = np.angle(root) / np.pi * F_nyquist_Hz
                new_roots = [(amp - 1) * F_nyquist_Hz + 1j * F_Hz]
            new_list.extend(new_roots)
        return new_list

    xfer_pre = fitter.xfer_fit

    poles = root_list_map(fitter.poles.fullplane, rtype="pole")
    zeros = root_list_map(fitter.zeros.fullplane, rtype="zero")

    # don't use the root mapper above since it does special things and can drop
    # overlays
    Zov, Pov, Kex = TFmath.ZtoS(
        (fitter.zeros_overlay.fullplane, fitter.poles_overlay.fullplane, 1),
        F_nyquist_Hz=fitter.F_nyquist_Hz,
    )

    ZPKrep = representations.ZPKwData(
        parent=fitter.ZPKrep,
        poles=poles,
        zeros=zeros,
        poles_overlay=Zov,
        zeros_overlay=Pov,
        gain=Kex * fitter.ZPKrep.gain,
    )

    xfer_post = ZPKrep.xfer_fit

    gain_scale = np.median(abs(xfer_post / xfer_pre))
    gain = fitter.gain / gain_scale * op.gain_flip

    ZPKrep.gain = gain
    return ZPKrep



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np
import cmath
from functools import reduce


def domain_sort(X, *Y):
    X = np.asarray(X)
    if not np.all(X[:-1] <= X[1:]):
        sort_idxs = np.argsort(X)
        X = X[sort_idxs]
        output = [X]
        for y in Y:
            if y is None:
                output.append(None)
            else:
                y = np.asarray(y)
                if len(y) == 1:
                    output.append(y)
                else:
                    output.append(y[sort_idxs])
    else:
        output = [X]
        output.extend(Y)
    return output


def broadcast_arrays_none(*args):
    arrm = []
    for arg in args:
        if arg is None:
            continue
        arrm.append(arg)
    arrm = np.broadcast_arrays(*arrm)
    ret = []
    idx = 0
    for arg in args:
        if arg is None:
            ret.append(None)
            continue
        ret.append(arrm[idx])
        idx += 1
    return ret


def select_through_none(select, *args):
    if select is None:
        return args
    aret = []
    for arg in args:
        if arg is None:
            aret.append(None)
            continue
        aret.append(arg[select])
    return aret


def interval_limit(X_min, X_max, X, *Y):
    X = np.asarray(X)
    X_idx = np.where((X >= X_min) & (X <= X_max))
    return (X[X_idx],) + tuple(np.asarray(y)[X_idx] for y in Y)


def masked_argsort(m_array):
    """
    Runs argsort on a masked array and only returns the argsort of the unmasked items
    """
    return np.argsort(m_array)[: sum(~m_array.mask)]


def continuous_phase(data, op_idx=0, sep=(1.01) * np.pi, deg=False, shiftmod=2):
    raw_angle = np.angle(data)
    diff = np.diff(raw_angle)
    sep = abs(sep)
    where_up = list(np.where(diff > sep)[0])
    where_down = list(np.where(diff < -sep)[0])
    value_mods = []
    shift = 0

    def shift_mod(val):
        return ((shiftmod + val) % (2 * shiftmod)) - shiftmod

    while True:
        if where_up and where_down:
            if where_up[-1] > where_down[-1]:
                shift = shift_mod(shift - 1)
                where = where_up.pop()
            else:
                shift = shift_mod(shift + 1)
                where = where_down.pop()
        elif where_up:
            shift = shift_mod(shift - 1)
            where = where_up.pop()
        elif where_down:
            shift = shift_mod(shift + 1)
            where = where_down.pop()
        else:
            break
        value_mods.append((where + 1, shift * 2 * np.pi))

    if not value_mods:
        if np.average(raw_angle) < -np.pi / 4:
            raw_angle += np.pi * 2
        if deg:
            raw_angle *= 180.0 / np.pi
        return raw_angle

    full_shift = np.empty_like(raw_angle)
    last_where, shift = value_mods.pop()
    full_shift[0:last_where] = shift
    while value_mods:
        new_where, shift = value_mods.pop()
        full_shift[last_where:new_where] = shift
        last_where = new_where
    full_shift[last_where:] = 0

    raw_angle -= full_shift
    raw_angle += full_shift[op_idx]
    while raw_angle[op_idx] < -np.pi:
        raw_angle += 2 * np.pi
    while raw_angle[op_idx] > np.pi:
        raw_angle -= 2 * np.pi

    median = np.sort(raw_angle)[len(raw_angle) / 2]
    if median < -np.pi / 4:
        raw_angle += np.pi * 2

    if deg:
        raw_angle *= 180.0 / np.pi
    return raw_angle


def logspaced(lower, upper, n_points):
    """
    Not very smart about preserving the number of points with a discontiguous interval set
    """
    log_lower = np.log(lower)
    log_upper = np.log(upper)
    return np.exp(np.linspace(log_lower, log_upper, n_points))


def common_type(nd_array):
    nd_flat = np.asanyarray(nd_array).flatten()
    return reduce(type_reduce, nd_flat, nd_flat[0].__class__)


def type_reduce(type_A, obj_B):
    if type_A is None or obj_B is None:
        return None
    if isinstance(obj_B, type_A):
        return type_A
    if issubclass(type_A, obj_B.__class__):
        return obj_B.__class__
    return None


def argsort(array):
    """
    Highly efficient argsort for pure python, this is also good for
    arrays where you only want the sort in the first dimesion
    """
    return sorted(list(range(len(array))), key=array.__getitem__)


def mag_phase_signed(v, deg=True):
    ang = (np.angle(v, deg=False) + np.pi * 9.0 / 4) % np.pi - np.pi / 4.0
    mag = v * np.exp(-1j * ang)
    if deg:
        ang = 180 / np.pi * ang
    return np.real(mag), ang


def group_delay(F, data, mult=3e8):
    dang = np.convolve([1, -1], np.angle(data), mode="valid")
    dang[dang > 1 * np.pi] -= 2 * np.pi
    dang[dang < -1 * np.pi] += 2 * np.pi
    dF = np.convolve([1, -1], F, mode="valid")
    return F[-len(dang) :], mult * dang / dF


def first_non_NaN(arr):
    idx_lower = 0
    idx_upper = len(arr)
    N = 1
    if not cmath.isnan(arr[0]):
        return 0
    while idx_lower + N < idx_upper:
        if not cmath.isnan(arr[idx_lower + N]):
            if N == 1:
                return idx_lower + 1
            else:
                idx_lower = idx_lower + N / 2
                idx_upper = idx_lower + N
                N = 1
        else:
            N *= 2
    return idx_upper


def search_local_sorted_orig(arr_x, arr_y, val_x_start, val_y):
    idx_start = np.searchsorted(arr_x, val_x_start)
    dval_y_start = arr_y[idx_start + 1] - arr_y[idx_start]
    idx_upper = idx_start
    idx_lower = idx_start
    if dval_y_start > 0:
        prev = arr_y[idx_start]
        while True:
            new = arr_y[idx_upper]
            if new < prev:
                break
            prev = new
            idx_upper += 1
            if idx_upper == len(arr_x):
                break
        prev = arr_y[idx_start]
        while True:
            new = arr_y[idx_lower]
            if new > prev:
                break
            prev = new
            idx_lower -= 1
            if idx_lower == -1:
                break
        idx_lower += 1
        idx_offset = np.searchsorted(arr_y[idx_lower:idx_upper], val_y)
    else:
        prev = arr_y[idx_start]
        while True:
            new = arr_y[idx_upper]
            if new > prev:
                break
            prev = new
            idx_upper += 1
            if idx_upper == len(arr_x):
                break
        prev = arr_y[idx_start]
        while True:
            new = arr_y[idx_lower]
            if new < prev:
                break
            prev = new
            idx_lower -= 1
            if idx_lower == -1:
                break
        idx_lower += 1
        idx_offset = -1 - np.searchsorted(arr_y[idx_lower:idx_upper][::-1], val_y)

    idx = idx_lower + idx_offset
    sub_idx = (val_y - arr_y[idx]) / (arr_y[idx + 1] - arr_y[idx])
    frac_x = arr_x[idx] + sub_idx * (arr_x[idx + 1] - arr_x[idx])
    return frac_x, idx, sub_idx


def search_local_sorted(arr_x, arr_y, val_x_start, val_y):
    idx_start = np.searchsorted(arr_x, val_x_start)
    dval_y = arr_y[1:] > arr_y[:-1]
    ddval_y = dval_y[1:] ^ dval_y[:-1]
    idx_convex = np.concatenate([[0], np.nonzero(ddval_y)[0], [len(arr_x)]])
    idx_split = np.searchsorted(idx_convex, idx_start)
    idx_lower = idx_convex[idx_split - 1]
    idx_upper = idx_convex[idx_split]

    if arr_y[idx_upper - 1] > arr_y[idx_lower]:
        idx_offset = np.searchsorted(arr_y[idx_lower:idx_upper], val_y)
    else:
        idx_offset = -1 - np.searchsorted(arr_y[idx_lower:idx_upper][::-1], val_y)

    idx = idx_lower + idx_offset
    sub_idx = (val_y - arr_y[idx]) / (arr_y[idx + 1] - arr_y[idx])
    frac_x = arr_x[idx] + sub_idx * (arr_x[idx + 1] - arr_x[idx])
    return frac_x, idx, sub_idx


def generate_sections(barray, reconnect_length=None):
    Dbarray = barray[1:] ^ barray[:-1]
    args = np.argwhere(Dbarray).T[0, :] + 1
    pargs = []
    if barray[0]:
        pargs.append([0])
    pargs.append(args)
    if barray[-1]:
        pargs.append([len(barray) - 1])
    if len(pargs) > 1:
        args = np.concatenate(pargs)
    assert len(args) % 2 == 0
    sections = list(zip(args[::2], args[1::2]))
    if (len(sections) > 0) and (reconnect_length is not None):
        disconnects = [sections[0][0]]
        for idx in range(1, len(sections)):
            _, eidx = sections[idx - 1]
            sidx, _ = sections[idx]
            if (sidx - eidx) > reconnect_length:
                disconnects.append(eidx)
                disconnects.append(sidx)
        disconnects.append(sections[-1][-1])
        sections = list(zip(disconnects[0::2], disconnects[1::2]))
    return sections


def generate_antisections(idx_start, idx_end, sections):
    if not sections:
        return [(idx_start, idx_end)]
    disconnects = []
    for section in sections:
        disconnects.extend(section)

    if disconnects[0] == idx_start:
        disconnects = disconnects[1:]
    else:
        disconnects.insert(0, idx_start)

    if disconnects[-1] == idx_end:
        disconnects = disconnects[:-1]
    else:
        disconnects.append(idx_end)
    return list(zip(disconnects[0::2], disconnects[1::2]))


def loglinspaced(lower, upper, n_points, frac=0.75):
    """
    Not very smart about preserving the number of points with a discontiguous interval set
    """
    n_log = int(n_points * frac)

    log_lower = np.log(lower)
    mid = (upper - lower) / 2 + lower
    mid_max = upper
    mid_min = lower
    for n in range(10):
        log_mid = np.log(mid)
        log_spacing = (1 - np.exp(-(log_mid - log_lower) / n_log)) * mid
        lin_spacing = (upper - mid) / (n_points - n_log)
        if log_spacing > lin_spacing:
            mid_max = mid
            mid = (mid - mid_min) / 2
        else:
            mid_min = mid
            mid = (mid_max - mid) / 2 + mid
    log_mid = np.log(mid)
    lin_spacing = (upper - mid) / (n_points - n_log)
    # print("SPACINGS", log_spacing, lin_spacing)
    return np.concatenate(
        [
            np.exp(np.linspace(log_lower, log_mid, n_points)),
            np.linspace(mid + lin_spacing, upper, n_points - n_log),
        ]
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
Some utilities to help with dataset creation and annotation
"""

import os
import os.path as path
import contextlib

from wield.iirrational import (
    fitters_rational,
    fitters_ZPK,
    representations,
)

# import matplotlib.pyplot as plt
from wield.iirrational import plots
from wield.iirrational.annotate.exporters import npm_markdown_pdf


@contextlib.contextmanager
def plot_on_assert(testfile, request, fitter, plot_anyway=False):
    if isinstance(request, (str)):
        rname = request
    else:
        rname = request.node.name
    fname = path.join(path.split(testfile)[0], "plots/{}.pdf".format(rname))

    def plot():
        if isinstance(
            fitter,
            (
                fitters_rational.DataFilterBase,
                fitters_ZPK.MultiReprFilterBase,
                representations.ZPKwData,
            ),
        ):
            axB = plots.plot_fitter_flag(fitter, fname=fname)
        else:
            axB = fitter()
            axB.save(fname)

    try:
        yield
    except AssertionError:
        plot()
        raise

    if plot_anyway:
        plot()
    else:
        try:
            os.remove(fname)
        except OSError:
            pass
    return


@contextlib.contextmanager
def digest_on_assert(testfile, request, aid, plot_anyway=False):
    if isinstance(request, (str)):
        rname = request
    else:
        rname = request.node.name
    fname = path.join(path.split(testfile)[0], "plots/{}/".format(rname))

    def plot():
        aid.digest_write(
            fname,
            plot_verbosity=9,
            clear_plots=True,
            plot_format="png",
            dpi=200,
            exporter=npm_markdown_pdf("digest.pdf"),
            MP_workers=1,
        )

    try:
        yield
    except AssertionError:
        plot()
        raise

    if plot_anyway:
        plot()
    else:
        try:
            # os.remove(fname)
            pass
        except OSError:
            pass
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import h5py
from wield.bunch.hdf_deep_bunch import HDFDeepBunch


def load_hdf5(fname):
    # with h5py.File(fname) as h5F:
    h5F = h5py.File(fname, "r")
    fdict = HDFDeepBunch(h5F, writeable=False)
    return fdict


def write_hdf5(fname, fdict):
    with h5py.File(fname, "w") as h5F:
        hdf = HDFDeepBunch(h5F, writeable=True)
        hdf.update_recursive(fdict)
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
Contains setup functions for test data, these are returned in a QuickBunch dictionary, with some annotation about the number of data sets
"""
import numpy as np
from wield.bunch import Bunch

from ..utilities.np import logspaced
from . import utilities


def meta_generator(
    Z,
    P,
    K,
    N=1000,
    SNR_default=10,
    delay_s=0,
    F_nyquist_Hz=None,
):
    def dataset_gen(
        sN=0, iN=0, SNR=None, invert=False, F_nyquist_Hz=F_nyquist_Hz, N=N, **kwargs
    ):
        if SNR is None:
            SNR = SNR_default

        if not invert:
            ZPK_s = (Z, P, K)
        else:
            ZPK_s = (P, Z, 1 / K)

        if sN == 0:
            F_Hz = np.linspace(0, 10, N)
        elif sN == 1:
            F_Hz = logspaced(0.01, 10, N)
        elif sN == 2:
            F_Hz = np.linspace(0, 100, N)
        else:
            F_Hz = logspaced(0.01, 100, N)

        return utilities.generator_autofill(
            F_Hz=F_Hz,
            SNR=SNR,
            F_nyquist_Hz=F_nyquist_Hz,
            bestfit_ZPK_s=ZPK_s,
            delay_s=delay_s,
            sN=sN,
            iN=iN,
        )

    return dataset_gen


datasets = Bunch()

datasets.simple0 = utilities.make_description(
    generator=meta_generator((), (-1,), 1),
    sets=4,
    instances=-1,
)

datasets.simple0E = utilities.make_description(
    generator=meta_generator(
        (),
        (-1,),
        1,
        N=100,
        SNR_default=None,
    ),
    sets=4,
    instances=-1,
)

datasets.simple1 = utilities.make_description(
    generator=meta_generator((-1, -1, -1, -1, -1), (-0.1, -0.1, -0.1), 1),
    sets=4,
    instances=-1,
)

datasets.simple1E = utilities.make_description(
    generator=meta_generator(
        (-1, -1, -1, -1, -1),
        (-0.1, -0.1, -0.1),
        1,
        N=30,
        SNR_default=None,
    ),
    sets=4,
    instances=-1,
)

datasets.delay = utilities.make_description(
    generator=meta_generator(
        (),
        (),
        1,
        delay_s=0.01,
    ),
    sets=4,
    instances=-1,
)

datasets.simple2 = utilities.make_description(
    generator=meta_generator(
        (-1, -1, -0.1 + 1.1j, -0.1 - 1.1j, -1),
        (
            -0.1,
            -0.1,
            -0.1 + 3j,
            -0.1 - 3j,
            -0.01 + 1j,
            -0.01 - 1j,
            -1,
        ),
        1,
    ),
    sets=4,
    instances=-1,
)

datasets.simple2_us = utilities.make_description(
    generator=meta_generator(
        (-1, -1, -0.1 + 1.1j, -0.1 - 1.1j, -1),
        (
            -0.1,
            -0.1,
            +0.1 + 3j,
            +0.1 - 3j,
            -0.01 + 1j,
            -0.01 - 1j,
            -1,
        ),
        1,
    ),
    sets=4,
    instances=-1,
)

datasets.simple2_nmd = utilities.make_description(
    generator=meta_generator(
        (-1, -1, +0.1 + 1.1j, +0.1 - 1.1j, -1),
        (
            -0.1,
            -0.1,
            -0.1 + 3j,
            -0.1 - 3j,
            -0.01 + 1j,
            -0.01 - 1j,
            -1,
        ),
        1,
    ),
    sets=4,
    instances=-1,
)

datasets.simple2_us2 = utilities.make_description(
    generator=meta_generator(
        (-1, -1, -0.1 + 1.1j, -0.1 - 1.1j, -1),
        (
            -0.1,
            -0.1,
            +0.1 + 3j,
            +0.1 - 3j,
            +0.01 + 1j,
            +0.01 - 1j,
            -1,
        ),
        1,
    ),
    sets=4,
    instances=-1,
)

datasets.simple2lownyquist = utilities.make_description(
    generator=meta_generator(
        (-1, -1, -0.1 + 1.1j, -0.1 - 1.1j, -1),
        (
            -0.1,
            -0.1,
            -0.1 + 3j,
            -0.1 - 3j,
            -0.01 + 1j,
            -0.01 - 1j,
            -1,
        ),
        1,
        F_nyquist_Hz=10,
    ),
    sets=4,
    instances=-1,
)

datasets.simple3 = utilities.make_description(
    generator=meta_generator(
        (
            -2,
            -2,
        ),
        (
            -0.1,
            -0.1,
            -0.1,
        ),
        1,
    ),
    sets=4,
    instances=-1,
)

datasets.simple4 = utilities.make_description(
    generator=meta_generator(
        (-0.1,),
        (
            -1,
            -1,
        ),
        1,
    ),
    sets=4,
    instances=-1,
)

datasets.testZPK4x = utilities.make_description(
    generator=meta_generator(
        (+0.1, +1 + 1j, +1 - 1j), (-0.5, -2 + 1j, -2 - 1j, -2 + 2j, -2 - 2j), 1
    ),
    sets=1,
    instances=-1,
)

datasets.testZPK4xHQ = utilities.make_description(
    # these are tuned to be right over 2x the log spacing
    # which is a special spacing used in the MRF code to not modify root locations
    generator=meta_generator(
        (),  # (-.1, -.1),
        (-0.03 + 2.0j, -0.03 - 2.0j),
        1,
        SNR_default=None,
    ),
    sets=1,
    instances=-1,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from .base import (
    CodingType,
    # Ipi,
    I2pi,
)


class CodingGain(CodingType):
    N_parameters = 1
    p_gain = 1

    def __init__(self):
        self.coding_id = 0
        return

    def clone(self, sys):
        new = self.__class__()
        new.update(gain=self.p_gain)
        return new

    def setup(self, gain=None):
        if gain is not None:
            self.p_gain = gain
        return

    @property
    def gain(self):
        return self.p_gain

    @gain.setter
    def gain(self, gain):
        self.p_gain = gain

    def update(self, gain):
        self.p_gain = gain

    def reduce(self):
        return [self.p_gain]

    def transfer(self):
        # gain term (can only be 1 and at the start)
        return self.p_gain

    def derivative(self):
        return [1]

    def roots(self):
        return []


class CodingDelay(CodingType):
    N_parameters = 1
    p_delay_s = 0

    def setup(self, delay_s=None):
        if delay_s is not None:
            self.p_delay_s = delay_s
        return

    @property
    def delay_s(self):
        return self.p_delay_s

    @delay_s.setter
    def delay_s(self, val):
        self.p_delay_s = val

    def update(self, delay):
        self.p_delay_s = delay

    def reduce(self):
        return [self.p_delay_s]

    def transfer(self):
        return np.exp(-I2pi * self.p_delay_s * self.sys.F_Hz)

    def derivative(self):
        return [-I2pi * self.sys.F_Hz * np.exp(-I2pi * self.p_delay_s * self.sys.F_Hz)]

    def roots(self):
        return []



# -*- coding: utf-8 -*-
"""
This code originally from module scipy.optimize.optimize
modified slightly to reduce dependencies.

SciPy license
Copyright © 2001, 2002 Enthought, Inc.
All rights reserved.

Copyright © 2003-2013 SciPy Developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer. Redistributions in binary
    form must reproduce the above copyright notice, this list of conditions and
    the following disclaimer in the documentation and/or other materials
    provided with the distribution. Neither the name of Enthought nor the names
    of the SciPy Developers may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import warnings
import numpy as np
import numpy as numpy
from wield.bunch import Bunch

try:
    from scipy.optimize import OptimizeWarning
except ImportError:

    class OptimizeWarning(UserWarning):
        pass


try:
    from scipy.optimize import OptimizeResult
except ImportError:
    OptimizeResult = Bunch


def neldermead(
    func,
    x0,
    args=(),
    callback=None,
    maxiter=None,
    maxfev=None,
    disp=False,
    return_all=False,
    initial_simplex=None,
    xatol=1e-4,
    fatol=1e-4,
    **unknown_options
):
    """
    Minimization of scalar function of one or more variables using the
    Nelder-Mead algorithm.

    Options
    -------
    disp : bool
        Set to True to print convergence messages.
    maxiter, maxfev : int
        Maximum allowed number of iterations and function evaluations.
        Will default to ``N*200``, where ``N`` is the number of
        variables, if neither `maxiter` or `maxfev` is set. If both
        `maxiter` and `maxfev` are set, minimization will stop at the
        first reached.
    initial_simplex : array_like of shape (N + 1, N)
        Initial simplex. If given, overrides `x0`.
        ``initial_simplex[j,:]`` should contain the coordinates of
        the j-th vertex of the ``N+1`` vertices in the simplex, where
        ``N`` is the dimension.
    xatol : float, optional
        Absolute error in xopt between iterations that is acceptable for
        convergence.
    fatol : number, optional
        Absolute error in func(xopt) between iterations that is acceptable for
        convergence.

    """
    if "ftol" in unknown_options:
        warnings.warn(
            "ftol is deprecated for Nelder-Mead,"
            " use fatol instead. If you specified both, only"
            " fatol is used.",
            DeprecationWarning,
        )
        if np.isclose(fatol, 1e-4) and not np.isclose(unknown_options["ftol"], 1e-4):
            # only ftol was probably specified, use it.
            fatol = unknown_options["ftol"]
        unknown_options.pop("ftol")
    if "xtol" in unknown_options:
        warnings.warn(
            "xtol is deprecated for Nelder-Mead,"
            " use xatol instead. If you specified both, only"
            " xatol is used.",
            DeprecationWarning,
        )
        if np.isclose(xatol, 1e-4) and not np.isclose(unknown_options["xtol"], 1e-4):
            # only xtol was probably specified, use it.
            xatol = unknown_options["xtol"]
        unknown_options.pop("xtol")

    _check_unknown_options(unknown_options)
    maxfun = maxfev
    retall = return_all

    fcalls, func = wrap_function(func, args)

    rho = 1
    chi = 2
    psi = 0.5
    sigma = 0.5
    nonzdelt = 0.05
    zdelt = 0.00025

    x0 = numpy.asfarray(x0).flatten()

    if initial_simplex is None:
        N = len(x0)

        sim = numpy.zeros((N + 1, N), dtype=x0.dtype)
        sim[0] = x0
        for k in range(N):
            y = numpy.array(x0, copy=True)
            if y[k] != 0:
                y[k] = (1 + nonzdelt) * y[k]
            else:
                y[k] = zdelt
            sim[k + 1] = y
    else:
        sim = np.asfarray(initial_simplex).copy()
        if sim.ndim != 2 or sim.shape[0] != sim.shape[1] + 1:
            raise ValueError("`initial_simplex` should be an array of shape (N+1,N)")
        if len(x0) != sim.shape[1]:
            raise ValueError("Size of `initial_simplex` is not consistent with `x0`")
        N = sim.shape[1]

    if retall:
        allvecs = [sim[0]]

    # If neither are set, then set both to default
    if maxiter is None and maxfun is None:
        maxiter = N * 200
        maxfun = N * 200
    elif maxiter is None:
        # Convert remaining Nones, to np.inf, unless the other is np.inf, in
        # which case use the default to avoid unbounded iteration
        if maxfun == np.inf:
            maxiter = N * 200
        else:
            maxiter = np.inf
    elif maxfun is None:
        if maxiter == np.inf:
            maxfun = N * 200
        else:
            maxfun = np.inf

    one2np1 = list(range(1, N + 1))
    fsim = numpy.zeros((N + 1,), float)

    for k in range(N + 1):
        fsim[k] = func(sim[k])

    ind = numpy.argsort(fsim)
    fsim = numpy.take(fsim, ind, 0)
    # sort so sim[0,:] has the lowest function value
    sim = numpy.take(sim, ind, 0)

    iterations = 1

    while fcalls[0] < maxfun and iterations < maxiter:
        if (
            numpy.max(numpy.ravel(numpy.abs(sim[1:] - sim[0]))) <= xatol
            and numpy.max(numpy.abs(fsim[0] - fsim[1:])) <= fatol
        ):
            break

        xbar = numpy.add.reduce(sim[:-1], 0) / N
        xr = (1 + rho) * xbar - rho * sim[-1]
        fxr = func(xr)
        doshrink = 0

        if fxr < fsim[0]:
            xe = (1 + rho * chi) * xbar - rho * chi * sim[-1]
            fxe = func(xe)

            if fxe < fxr:
                sim[-1] = xe
                fsim[-1] = fxe
            else:
                sim[-1] = xr
                fsim[-1] = fxr
        else:  # fsim[0] <= fxr
            if fxr < fsim[-2]:
                sim[-1] = xr
                fsim[-1] = fxr
            else:  # fxr >= fsim[-2]
                # Perform contraction
                if fxr < fsim[-1]:
                    xc = (1 + psi * rho) * xbar - psi * rho * sim[-1]
                    fxc = func(xc)

                    if fxc <= fxr:
                        sim[-1] = xc
                        fsim[-1] = fxc
                    else:
                        doshrink = 1
                else:
                    # Perform an inside contraction
                    xcc = (1 - psi) * xbar + psi * sim[-1]
                    fxcc = func(xcc)

                    if fxcc < fsim[-1]:
                        sim[-1] = xcc
                        fsim[-1] = fxcc
                    else:
                        doshrink = 1

                if doshrink:
                    for j in one2np1:
                        sim[j] = sim[0] + sigma * (sim[j] - sim[0])
                        fsim[j] = func(sim[j])

        ind = numpy.argsort(fsim)
        sim = numpy.take(sim, ind, 0)
        fsim = numpy.take(fsim, ind, 0)
        if callback is not None:
            callback(sim[0])
        iterations += 1
        if retall:
            allvecs.append(sim[0])

    x = sim[0]
    fval = numpy.min(fsim)
    warnflag = 0

    if fcalls[0] >= maxfun:
        warnflag = 1
        msg = _status_message["maxfev"]
        if disp:
            print("Warning: " + msg)
    elif iterations >= maxiter:
        warnflag = 2
        msg = _status_message["maxiter"]
        if disp:
            print("Warning: " + msg)
    else:
        msg = _status_message["success"]
        if disp:
            print(msg)
            print("         Current function value: %f" % fval)
            print("         Iterations: %d" % iterations)
            print("         Function evaluations: %d" % fcalls[0])

    result = OptimizeResult(
        fun=fval,
        nit=iterations,
        nfev=fcalls[0],
        status=warnflag,
        success=(warnflag == 0),
        message=msg,
        x=x,
        final_simplex=(sim, fsim),
    )
    if retall:
        result["allvecs"] = allvecs
    return result


def _check_unknown_options(unknown_options):
    if unknown_options:
        msg = ", ".join(map(str, unknown_options.keys()))
        # Stack level 4: this is called from _minimize_*, which is
        # called from another function in Scipy. Level 4 is the first
        # level in user code.
        warnings.warn("Unknown solver options: %s" % msg, OptimizeWarning, 4)


_status_message = {
    "success": "Optimization terminated successfully.",
    "maxfev": "Maximum number of function evaluations has been exceeded.",
    "maxiter": "Maximum number of iterations has been exceeded.",
    "pr_loss": "Desired error not necessarily achieved due to precision loss.",
}


def wrap_function(function, args):
    ncalls = [0]
    if function is None:
        return ncalls, None

    def function_wrapper(*wrapper_args):
        ncalls[0] += 1
        return function(*(wrapper_args + args))

    return ncalls, function_wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from wield.bunch.depbunch import (
    DepBunch,
    # Bunch,
    depB_property,
    NOARG,
)

from .. import representations


class DataFilterBase(DepBunch):
    RBalgo = representations.RBAlgorithms()
    root_constraint = RBalgo.root_constraints.mirror_real

    def __build__(self, _args=None, parent=NOARG, ZPKrep=NOARG, **kwargs):
        if _args is not None:
            raise RuntimeError("Only keyword arguments allowed")

        if parent is not NOARG and ZPKrep is NOARG:
            ZPKrep = parent.ZPKrep

        if ZPKrep is not NOARG:
            kwargs["parent"] = ZPKrep
        ZPKrep = representations.ZPKwData(**kwargs)

        super(DataFilterBase, self).__build__()
        self.ZPKrep = ZPKrep

    @depB_property
    def F_max_Hz(self):
        return np.max(self.F_Hz)

    @depB_property
    def X_grid(self):
        return self.phi_rep(self.F_Hz)

    def xfer_eval(self, F_Hz):
        # TODO, use ZPKrep for this...
        alt_filter = self.__class__(
            F_Hz=F_Hz,
            W=None,
            data=None,
            parent=self,
        )
        # providing the numerical scale will aid its precision
        # note! it should be computed in a more numerically stable way!
        # TODO use a ZPK computation for this
        return alt_filter.xfer_fit

    @depB_property
    def ZPKrep(self, rep=NOARG):
        if rep is NOARG:
            Z, lnGZ = self.phi_rep_native_lnG(self.zeros)
            P, lnGP = self.phi_rep_native_lnG(self.poles)
            Zov, lnGZov = self.phi_rep_native_lnG(self.zeros_overlay)
            Pov, lnGPov = self.phi_rep_native_lnG(self.poles_overlay)
            return representations.ZPKwData(
                data=self.data,
                W=self.W,
                F_Hz=self.F_Hz,
                zeros=Z,
                poles=P,
                zeros_overlay=Zov,
                poles_overlay=Pov,
                gain=self.gain * np.exp(-(lnGZ + lnGZov) + (lnGP + lnGPov)),
                delay_s=self.delay_s,
                F_nyquist_Hz=self.F_nyquist_Hz,
            )
        else:
            self.data = rep.data
            self.W = rep.W
            self.F_Hz = rep.F_Hz
            self.delay_s = rep.delay_s
            self.F_nyquist_Hz = rep.F_nyquist_Hz

            # must be assigned
            Z, lnGZ = self.phi_native_rep_lnG(rep.zeros)
            P, lnGP = self.phi_native_rep_lnG(rep.poles)
            Zov, lnGZov = self.phi_native_rep_lnG(rep.zeros_overlay)
            Pov, lnGPov = self.phi_native_rep_lnG(rep.poles_overlay)
            self.zeros = Z
            self.poles = P
            self.zeros_overlay = Zov
            self.poles_overlay = Pov
            self.gain = rep.gain * np.exp(-(lnGZ + lnGZov) + (lnGP + lnGPov))

            self.dependencies("gain")
            self.dependencies("zeros")
            self.dependencies("poles")
            self.dependencies("zeros_overlay")
            self.dependencies("poles_overlay")
            self.dependencies("data")
            self.dependencies("W")
            self.dependencies("delay_s")
            self.dependencies("F_Hz")
            self.dependencies("F_nyquist_Hz")
            return rep

    def phi_coeff_norm(self, pvec):
        return pvec[-1] * self.phi_gain_adj_pvec(pvec)

    def phi_gain_adj_pvec(self, pvec):
        return 1

    def phi_gain_adj_roots(self, rB):
        return 1

    def phi_rep(self, F_Hz):
        """
        function mapping data frequencies from frequency domain to fit domain
        """
        return 1j * F_Hz

    def phi_rep_native_lnG(self, rB):
        """
        function mapping roots to native domain (S) from fit representation domain
        outputs both the remapped roots, as well as the lnG adjustment
        """
        rB = self.RBalgo.expect(rB, self.RBalgo.root_constraints.mirror_real)
        return rB, 0

    def phi_native_rep_lnG(self, rB):
        """
        function mapping roots from native domain (S) to fit representation domain
        outputs both the remapped roots, as well as the lnG adjustment
        """
        rB = self.RBalgo.expect(rB, self.RBalgo.root_constraints.mirror_real)
        return rB, 0

    def phi_rep_Snative(self, rB):
        """
        Maps the internal representation to Snative, the output RootBunch can
        have a specialty constraint added.
        """
        rB = self.RBalgo.expect(rB, self.RBalgo.root_constraints.mirror_real)
        return rB

    def phi_Snative_rep(self, rB):
        """ """
        rB = self.RBalgo.expect(rB, self.RBalgo.root_constraints.mirror_real)
        return rB

    def _phasing_roots(self, roots):
        return 1

    def _phasing_vec(self, vec):
        return 1

    def _phasing_afit(self):
        return 1

    def _phasing_bfit(self):
        return 1

    @depB_property
    def order(self):
        return max(
            len(self.poles) + len(self.poles_overlay),
            len(self.zeros) + len(self.zeros_overlay),
        )

    @depB_property
    def order_sos(self):
        return (
            max(
                len(self.poles) + len(self.poles_overlay),
                len(self.zeros) + len(self.zeros_overlay),
            )
            + 1
        ) // 2

    @depB_property
    def order_relative(self):
        return (len(self.zeros) + len(self.zeros_overlay)) - (
            len(self.poles) + len(self.poles_overlay)
        )


class ZFilterBase(DataFilterBase):
    @depB_property
    def F_nyquist_Hz(self, val=NOARG):
        if val is NOARG:
            raise RuntimeError("Must specify F_nyquist_Hz before using class")
        if val is None:
            raise RuntimeError("F_nyquist_Hz must be positive for Z filters")
        return val

    @depB_property
    def Xn_grid(self):
        return self.phi_rep(-self.F_Hz)

    @depB_property
    def ZPKz(self):
        rep = self.ZPKrep
        return representations.ZPKCalc(
            tuple(rep.zeros.fullplane) + tuple(rep.zeros_overlay.fullplane),
            tuple(rep.poles.fullplane) + tuple(rep.poles_overlay.fullplane),
            self.gain,
            F_nyquist_Hz=self.F_nyquist_Hz,
        )

    def phi_gain_adj_roots(self, roots):
        return 1

    def phi_gain_adj_pvec(self, pvec):
        return 1

    def phi_rep(self, F_Hz):
        """
        function mapping roots from frequency domain to fit domain
        """
        return np.exp(1j * np.pi * F_Hz / self.F_nyquist_Hz)

    def phi_rep_native_lnG(self, rB):
        """
        function mapping roots to native domain (S) from fit representation domain
        outputs both the remapped roots, as well as the lnG adjustment
        """
        rB = self.RBalgo.expect(rB, self.RBalgo.root_constraints.mirror_real)
        return rB, 0

    def phi_native_rep_lnG(self, rB):
        """
        function mapping roots from native domain (S) to fit representation domain
        outputs both the remapped roots, as well as the lnG adjustment
        """
        rB = self.RBalgo.expect(rB, self.RBalgo.root_constraints.mirror_real)
        return rB, 0

    def phi_rep_Snative(self, rB):
        """
        Maps the internal representation to Snative, the output RootBunch can
        have a specialty constraint added.
        """
        rB.c
        return rB

    def phi_Snative_rep(self, rB):
        """ """
        return rB


class SFilterBase(DataFilterBase):
    def __init__(self, _args=None, F_nyquist_Hz=None, **kwargs):
        super(SFilterBase, self).__init__(_args, F_nyquist_Hz=F_nyquist_Hz, **kwargs)

    @depB_property
    def F_nyquist_Hz(self, val=NOARG):
        if val is NOARG:
            return None
        if val is not None:
            raise RuntimeError("F_nyquist_Hz must None for S filters")
        return val

    @depB_property
    def ZPKsf(self):
        rep = self.ZPKrep
        return representations.ZPKCalc(
            tuple(rep.zeros.fullplane) + tuple(rep.zeros_overlay.fullplane),
            tuple(rep.poles.fullplane) + tuple(rep.poles_overlay.fullplane),
            rep.gain,
            F_nyquist_Hz=rep.F_nyquist_Hz,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np

from ..codings_cmn import (
    CodingTypeZ,
    Ipi,
)


class CodingnlFnlBW(CodingTypeZ):
    N_parameters = 2
    unstable = False
    p_nlF_Hz = 0
    p_nl_BW = 0
    min_BW_Hz = 0

    def update(self, nlF_Hz, nl_BW):
        self.p_nlF_Hz = nlF_Hz
        self.p_nl_BW = nl_BW

    def reduce(self):
        return [
            self.p_nlF_Hz,
            self.p_nl_BW,
        ]

    def option_set(self, minimum_BW_Hz=None, **kwargs):
        super(CodingnlFnlBW, self).option_set(**kwargs)
        if minimum_BW_Hz is not None:
            self.min_BW_Hz = minimum_BW_Hz
        # TODO, should modify p_nl_BW for the new min_BW_Hz
        return

    def check_dist_limit(self, F_Hz=None, thresh=1):
        if self.sys.distance_limit_auto >= thresh:
            if F_Hz is None:
                self.min_BW_Hz = self.sys.distance_limit(self.F_Hz)
            else:
                self.min_BW_Hz = self.sys.distance_limit(F_Hz)
            return True
        return False

    def update_roots(self, r1):
        """ """
        # TODO should honor min_BW_Hz
        amp = abs(r1)
        F_Hz = abs(np.angle(r1) / np.pi * self.sys.F_nyquist_Hz)
        self.check_dist_limit(F_Hz=F_Hz, thresh=1)

        ret = True
        self.F_cutoff_Hz = self.sys.F_cutoff_Hz
        relF_Hz = F_Hz / self.F_cutoff_Hz
        if relF_Hz >= 1:
            relF_Hz = 0.999
            ret = False
        elif relF_Hz <= -1:
            relF_Hz = 0.001
            ret = False
        self.p_nlF_Hz = relF_Hz / (1 - relF_Hz ** 2) ** 0.5

        BW_Hz = (amp - 1) * self.sys.F_nyquist_Hz
        if BW_Hz <= 0:
            self.unstable = False
            BWp = -BW_Hz
        else:
            self.unstable = True
            BWp = BW_Hz

        if BWp > self.min_BW_Hz:
            self.p_nl_BW = (BWp - self.min_BW_Hz) ** 0.5
        else:
            self.p_nl_BW = self.min_BW_Hz ** 0.5
            ret = False
        return ret

    @property
    def gain_effect(self):
        # always 1 since it is complex
        if not self.unstable:
            return 1
        else:
            return 1

    @property
    def relF_Hz(self):
        val = self.p_nlF_Hz / (1.0 + self.p_nlF_Hz ** 2) ** 0.5
        return val

    @property
    def F_Hz(self):
        return self.F_cutoff_Hz * self.relF_Hz

    @property
    def BW(self):
        if not self.unstable:
            BW = -self.p_nl_BW ** 2
        else:
            BW = self.p_nl_BW ** 2
        return BW

    def transfer(self):
        self.check_dist_limit(thresh=2)
        # frequency, logarithmic amplitude
        amp = 1 + self.BW / self.sys.F_nyquist_Hz

        r = amp * np.cos(np.pi * self.F_Hz / self.sys.F_nyquist_Hz)
        Xn = self.sys.Xzn_grid
        Xnsq = self.sys.Xzn_grid_sq
        return (amp * amp) * Xnsq - 2 * Xn * r + 1

    def derivative(self):
        self.check_dist_limit(thresh=2)
        # real/imaginary part of root
        if not self.unstable:
            DampDl = -(2 * self.p_nl_BW) / self.sys.F_nyquist_Hz
            amp = 1 - self.p_nl_BW ** 2 / self.sys.F_nyquist_Hz
        else:
            DampDl = +(2 * self.p_nl_BW) / self.sys.F_nyquist_Hz
            amp = 1 + self.p_nl_BW ** 2 / self.sys.F_nyquist_Hz

        Fsqp5 = (1.0 + self.p_nlF_Hz ** 2) ** 0.5
        relF_Hz = self.p_nlF_Hz / Fsqp5
        DrelFHzDp = 1 / Fsqp5 - self.p_nlF_Hz ** 2 / (1.0 + self.p_nlF_Hz ** 2) ** 1.5

        F_nyquist_Hz = self.sys.F_nyquist_Hz
        Xn = self.sys.Xzn_grid
        Xnsq = self.sys.Xzn_grid_sq
        rcos = np.cos(np.pi * self.F_cutoff_Hz * relF_Hz / F_nyquist_Hz)
        rsin = np.sin(np.pi * self.F_cutoff_Hz * relF_Hz / F_nyquist_Hz)
        return [
            DrelFHzDp
            * (2 * amp * np.pi * self.F_cutoff_Hz / self.sys.F_nyquist_Hz)
            * (Xn * rsin),
            ((2 * amp * DampDl) * Xnsq - Xn * (2 * rcos * DampDl)),
        ]

    def roots_c(self):
        amp = 1 + self.BW / self.sys.F_nyquist_Hz
        # real/imaginary part of root
        return [
            amp
            * np.exp(Ipi * abs(self.relF_Hz) * self.F_cutoff_Hz / self.sys.F_nyquist_Hz)
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ..codings_cmn import (
    CodingTypeZ,
)


class CodingRI(CodingTypeZ):
    N_parameters = 2
    p_real = 0
    p_imag = 0

    def update(self, real, imag):
        self.p_real = real
        self.p_imag = imag

    def reduce(self):
        return [
            self.p_real,
            self.p_imag,
        ]

    @property
    def gain_effect(self):
        if self.p_real ** 2 + self.p_imag ** 2 > 1:
            return -1
        return 1

    def update_roots(self, r1, r2=None):
        """
        r2, may be unspecified, in which case it is assumed to be nothing, if r1 is real, or otherwise the conjugate of r1
        """
        # TODO, check r2
        self.p_real = r1.real
        self.p_imag = r1.imag
        return

    def transfer(self):
        # real/imaginary part of root
        r, i = self.p_real, self.p_imag
        Xn = self.sys.Xzn_grid
        Xnsq = self.sys.Xzn_grid_sq
        return (r * r + i * i) * Xnsq - 2 * Xn * r + 1

    def derivative(self):
        # real/imaginary part of root
        r, i = self.p_real, self.p_imag
        Xn = self.sys.Xzn_grid
        Xnsq = self.sys.Xzn_grid_sq
        return [
            2 * (r * Xnsq - Xn),
            (2 * i) * Xnsq,
        ]

    def roots_c(self):
        # real/imaginary part of root
        return [self.p_real + 1j * abs(self.p_imag)]
